/*
 * Tilt Click driver for MikroBusNet
 * 
 * Version 1.0 -Initial version coded by Stephen Cardinale
 * 
 * References needed :
 *  Microsoft.SPOT.Hardware
 *  Microsoft.SPOT.Native
 *  MikroBusNet
 *  mscorlib
 *  
 * Copyright 2015 MikroBus.Net
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
 * either express or implied. See the License for the specific language governing permissions and limitations under the License.
 */
using System;
using System.Reflection;
using System.Threading;
using MBN.Enums;
using MBN.Exceptions;
using Microsoft.SPOT.Hardware;

namespace MBN.Modules
{
    /// <summary>
    /// MiKroBusNet NetMF Driver for a MikroE Tilt Click.
    /// <para><b>The Tilt Click is a Generic Module.</b></para>
    /// <para><b>Pins used :</b> Pwm, Int</para>
    /// <para><b>References required:</b> MikroBus.Net, Microsoft.SPOT.Native, Microsoft.Spot.Hardware, mscorlib</para>
    /// </summary>
    /// <example>Example usage:
    /// <code language = "C#">
    /// using System;
    /// using System.Threading;
    /// using MBN;
    /// using MBN.Modules;
    /// using Microsoft.SPOT;
    ///
    /// namespace Examples
    /// {
    ///     public class Program
    ///     {
    ///         private static TiltClick tilt;
    /// 
    ///         public static void Main2()
    ///         {
    ///             tilt = new TiltClick(Hardware.SocketTwo, false)
    ///             tilt.SamplingFrequency = new TimeSpan(0, 0, 0, 0, 500),
    ///
    ///             tilt.TiltChanged += tilt_TiltChanged;
    ///             tilt.StartContinuousMeasurements();
    ///             Thread.Sleep(-1);
    ///         }
    ///
    ///         static void tilt_TiltChanged(object sender, TiltClick.Direction direction)
    ///         {
    ///             Debug.Print("Tilt direction - " + (direction == TiltClick.Direction.Right ? "Right" : direction == TiltClick.Direction.Left ? "Left" : direction == TiltClick.Direction.Forward ? "Forward" : "Backward"));
    ///         }
    ///     }
    /// }
    /// </code>
    /// <code language = "VB">
    /// Option Explicit On
    /// Option Strict On
    ///
    /// Imports System.Threading
    /// Imports MBN
    /// Imports MBN.Modules
    /// Imports Microsoft.SPOT
    ///
    /// Namespace Examples
    /// 	Public Class Program
    /// 		Private Shared tilt As TiltClick
    ///
    /// 		Public Shared Sub Main2()
    ///
    /// 			tilt = New TiltClick(Hardware.SocketTwo, False)
    ///             tilt.SamplingFrequency = New TimeSpan(0, 0, 0, 0, 500)
    ///
    /// 			AddHandler tilt.TiltChanged, AddressOf tilt_TiltChanged
    ///
    /// 			tilt.StartContinuousMeasurements()
    /// 			Thread.Sleep(Timeout.Infinite1)
    /// 		End Sub
    ///
    /// 		Private Shared Sub tilt_TiltChanged(sender As Object, direction As TiltClick.Direction)
    /// 			Debug.Print("Tilt direction - " <![CDATA[&]]> (If(direction = TiltClick.Direction.Right, "Right", If(direction = TiltClick.Direction.Left, "Left", If(direction = TiltClick.Direction.Forward, "Forward", "Backward")))))
    /// 		End Sub
    /// 	End Class
    /// End Namespace
    /// </code>
    /// </example>
    public class TiltClick : IDriver
    {
        #region Fields

        private static InputPort vOut1;
        private static InputPort vOut2;
        private static InterruptPort vOut1Int;
        private static InterruptPort vOut2Int;

        private static SensorOrientation _orientation = SensorOrientation.Normal;
        private static Boolean _usePolling = true;
        private static Timer _pollingTimer;
        private static bool _isPolling;
        private TimeSpan _samplingFrequency = new TimeSpan(0, 0, 0, 5);
        private TimeSpan _debounceInterval = TimeSpan.FromTicks(375000);
        private static DateTime _lastInterrupt;


        #endregion
        
        #region ENUMS
        
        /// <summary>
        /// The direction of the tilt. With <see cref="SensorOrientation.Normal"/> the Forward is located closest to the RPI-1035 sensor (furthest away from the Ground Pins).
        /// </summary>
        public enum Direction
        {
            /// <summary>
            /// The Tilt click is tilted to the left (left is up).
            /// </summary>
            Left = 0,
            /// <summary>
            /// The Tilt click is tilted to the right (right is up).
            /// </summary>
            Right = 1,
            /// <summary>
            /// The Tilt click is tilted forward the front (front is up).
            /// </summary>
            Forward = 3,
            /// <summary>
            /// The Tilt click is tilted backward (back is up).
            /// </summary>
            Backward = 4
        }

        /// <summary>
        /// Sensor orientation. Useful to change the perspective as the Sockets 1 and 2 on the Quail board are positioned 180 orientation from Sockets 3 and 4. 
        /// </summary>
        public enum SensorOrientation
        {
            /// <summary>
            /// Normal orientation, as viewed, the RPI-1035 sensor is located away from you.
            /// </summary>
            Normal = 0,
            /// <summary>
            /// Rotated orientation, as viewed, the RPI-1035 sensor is located closest to you.
            /// </summary>
            Rotate180 = 1
        }

        #endregion
        
        #region CTOR

        /// <summary>
        /// Initializes a new instance of the <see cref="TiltClick"/> class.
        /// </summary>
        /// <param name="socket">The <see cref="Hardware.Socket"/> that the Tilt Click is plugged into.</param>
        /// <param name="usePolling">Set to true to use Polling and false to use Interrupt driven approach. </param>
        public TiltClick(Hardware.Socket socket, Boolean usePolling)
        {
            try
            {
                Hardware.CheckPins(socket, socket.Int, socket.Pwm);

                _usePolling = usePolling;
                
                if (usePolling)
                {
                    vOut1 = new InputPort(socket.Pwm, false, Port.ResistorMode.PullDown);
                    vOut2 = new InputPort(socket.Int, true, Port.ResistorMode.PullDown);
                    _pollingTimer = new Timer(UpdateReadings, null, Timeout.Infinite, Timeout.Infinite);
                }
                else
                {
                    vOut1Int = new InterruptPort(socket.Pwm, true, Port.ResistorMode.PullUp, Port.InterruptMode.InterruptEdgeBoth);
                    vOut2Int = new InterruptPort(socket.Int, true, Port.ResistorMode.PullUp, Port.InterruptMode.InterruptEdgeBoth);
                    vOut1Int.OnInterrupt += vOut_OnInterrupt;
                    vOut2Int.OnInterrupt += vOut_OnInterrupt;
                    _lastInterrupt = DateTime.Now;
                }
            }
            catch (PinInUseException ex)
            {
                throw new PinInUseException(ex.Message);
            }
        }

        #endregion

        #region Private Methods

        private void vOut_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            if (time.Subtract(_lastInterrupt) <= _debounceInterval) return; // Do not want to fire too many false events.

            var direction = Direction.Forward;

            if ((vOut2Int.Read()) && (vOut1Int.Read() == false))
            {
                direction = (_orientation == SensorOrientation.Normal ? Direction.Right : Direction.Left);
            }
            if ((vOut2Int.Read()) == false && vOut1Int.Read())
            {
                direction = _orientation == SensorOrientation.Normal ? Direction.Left : Direction.Right;
            }
            if ((vOut2Int.Read()) && (vOut1Int.Read()))
            {
                direction = _orientation == SensorOrientation.Normal ? Direction.Backward : Direction.Forward;
            }
            if ((vOut2Int.Read() == false) && (vOut1Int.Read() == false))
            {
                direction = _orientation == SensorOrientation.Normal ? Direction.Forward : Direction.Backward;
            }

            _lastInterrupt = time;

            TiltChanged(this, direction);
        }
        
        private void UpdateReadings(object state)
        {
            if (!_isPolling) return;
            var tilt = GetTiltDirection();
            TiltChanged(this, tilt);
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Reads the Tilt click direction of tilt.
        /// </summary>
        /// <returns>The <see cref="Direction"/> of the tilt.</returns>
        /// <example>Example usage:
        /// <code language = "C#">
        /// var direction = tilt.GetTiltDirection();
        /// Debug.Print("Direction: " + (direction == TiltClick.Direction.Right ? "Right" : direction == TiltClick.Direction.Left ? "Left" : direction == TiltClick.Direction.Forward ? "Forward" : "Backward"));
        /// </code>
        /// <code language = "VB">
        /// Dim direction = tilt.GetTiltDirection()
        /// Debug.Print("Direction: " <![CDATA[&]]> (If(direction = TiltClick.Direction.Right, "Right", If(direction = TiltClick.Direction.Left, "Left", If(direction = TiltClick.Direction.Forward, "Forward", "Backward")))))
        /// </code>
        /// </example>
        public Direction GetTiltDirection()
        {
            if (vOut2.Read() && (vOut1.Read() == false))
            {
                return _orientation == SensorOrientation.Normal ? Direction.Right : Direction.Left;
            }
            if (vOut2.Read() == false && vOut1.Read())
            {
                return _orientation == SensorOrientation.Normal ? Direction.Left : Direction.Right;
            }
            if (vOut2.Read() && (vOut1.Read()))
            {
                return _orientation == SensorOrientation.Normal ? Direction.Backward : Direction.Forward;
            }
            return _orientation == SensorOrientation.Normal ? Direction.Forward : Direction.Backward;
        }

        /// <summary>
        /// Starts continuous polling of the Tilt Click based on the time frequency set in the <see cref="SamplingFrequency"/> Property.
        /// If the SamplingFrequency Property is not set, it will default to 5 seconds.
        /// </summary>
        /// <remarks>Tilt will be returned by raising the <see cref="TiltChanged"/> Event.
        /// </remarks>
        /// <example>Example usage:
        /// <code language = "C#">
        /// _tiltClick.StartContinuousMeasurements();
        /// </code>
        /// <code language = "VB">
        /// _tiltClick.StartContinuousMeasurements()
        /// </code>
        /// </example>
        public void StartContinuousMeasurements()
        {
            if (_isPolling || _usePolling == false) return;
            _pollingTimer.Change(TimeSpan.Zero, _samplingFrequency == TimeSpan.Zero ? new TimeSpan(0, 0, 0, 15) : _samplingFrequency); //Default is every 15 seconds.
            _isPolling = true;
        }

        /// <summary>
        ///     Stops continuous polling of the Tilt Click for tilt direction.
        /// </summary>
        /// <example>Example usage:
        /// <code language = "C#">
        /// _tiltClick.StopContinuousMeasurements();
        /// </code>
        /// <code language = "VB">
        /// _tiltClick.StopContinuousMeasurements()
        /// </code>
        /// </example>
        public void StopContinuousMeasurements()
        {
            if (!_isPolling) return;
            _pollingTimer.Change(TimeSpan.Zero, TimeSpan.MaxValue);
            _isPolling = false;
        }

        /// <summary>
        /// Resets the module
        /// </summary>
        /// <param name="resetMode">The reset mode :
        /// <para>SOFT reset : generally by sending a software command to the chip</para><para>HARD reset : generally by activating a special chip's pin</para></param>
        /// <example>None provided as this module does not support <see cref="ResetModes"/>.
        /// </example>
        /// <exception cref="System.NotImplementedException">Thrown because this module has no reset feature.</exception>
        public bool Reset(ResetModes resetMode)
        {
            throw new NotImplementedException();
        }

        #endregion  
      
        #region Public Properties
		
        /// <summary>
        /// Gets or sets the orientation of the of the Tilt Click.
        /// </summary>
        /// <example>Example 
        /// <code language = "C#">
        /// _tilt.Orientation = SensorOrientation.Rotate180;
        /// </code>
        /// <code language = "VB">
        /// _tilt.Orientation = SensorOrientation.Rotate180
        /// </code>
        /// </example>
        public SensorOrientation Orientation
        {
            get { return _orientation; }
            set { _orientation = value; }
        }

        /// <summary>
        /// Gets or sets the debounce interval for stopping multiple <see cref="TiltChanged"/> events from firing when using the Interrupt (non-polling) mode.
        /// <p>Defaults to 37.5 milliseconds. Increase this value accordingly if you are receiving numerous false TiltChanged events.</p>
        /// </summary>
        /// <example>Example 
        /// <code language = "C#">
        /// tilt.DebounceInterval = new TimeSpan(0, 0, 0, 0, 50);
        /// </code>
        /// <code language = "VB">
        /// tilt.DebounceInterval = new TimeSpan(0, 0, 0, 0, 50)
        /// </code>
        /// </example>
        public TimeSpan DebounceInterval
        {
            get { return _debounceInterval; }
            set { _debounceInterval = value; }
        }

        /// <summary>
        ///     Sets the TimeSpan interval in which the Tilt Click will be polled to return tilt direction.
        /// </summary>
        /// <example>Example usage:
        /// <code language = "C#">
        /// _tiltClick.SamplingFrequency = new TimeSpan(0, 0, 0, 2);
        /// </code>
        /// <code language = "VB">
        /// _tiltClick.SamplingFrequency = new TimeSpan(0, 0, 0, 2)
        /// </code>
        /// </example>
        public TimeSpan SamplingFrequency
        {
            get { return _samplingFrequency; }
            set
            {
                if (value < new TimeSpan(2500000)) value = new TimeSpan(2500000); // Do not sample faster than the conversion time plus a few milliseconds.
                _samplingFrequency = value;
            }
        }
        
        /// <summary>
        /// Gets or sets the power mode.
        /// </summary>
        /// <example> This sample shows how to use the PowerMode property.
        /// </example>
        /// <value>
        /// The current power mode of the module.
        /// </value>
        /// <exception cref="System.NotImplementedException">Thrown if the property is set and the module doesn't support power modes.</exception>
        /// <example>Example usage: None provided as this module does not support <see cref="PowerModes"/>.
        /// <code language = "C#">
        /// // This module does not support Power Mode selection.
        /// </code>
        /// <code language = "VB">
        /// ' This module does not support Power Mode selection.
        /// </code>
        /// </example>
        public PowerModes PowerMode
        {
            get { return PowerModes.On; }
            set
            { 
                throw new NotImplementedException();
            }
        }

        /// <summary>
        /// Gets the driver version.
        /// </summary>
        /// <example> This sample shows how to use the DriverVersion property.
        /// <code language="C#">
        /// Debug.Print ("Current driver version : "+ _tiltClick.DriverVersion);
        /// </code>
        /// <code language = "VB">
        /// Debug.Print ("Current driver version : " <![CDATA[&]]> _tiltClick.DriverVersion)
        /// </code>
        /// </example>
        /// <value>
        /// The driver version.
        /// </value>
        public Version DriverVersion
        {
            get { return Assembly.GetExecutingAssembly().GetName().Version; }
        }
        
	    #endregion

        #region Events

        /// <summary>
        ///     This event is raised when the Tilt Sensor completes a measurement in continuous measurement mode.
        /// </summary>
        public event TiltChangedEventHandler TiltChanged;

        /// <summary>
        ///     Represents the delegate that is used for the <see cref="TiltChanged" />.
        /// </summary>
        /// <param name="sender">The DHT22 Sensor that raised </param>
        /// <param name="direction">The <see cref="Direction"/> of the tilt.</param>
        public delegate void TiltChangedEventHandler(object sender, Direction direction);

        #endregion

    }
}
